@@ -20,21 +20,30 @@
 
 #include "config.h"
 
+#include <gio/gio.h>
 #include <glib/gi18n.h>
-#include <libsecret/secret.h>
 #include <stdio.h>
 
-#include "grd-schemas.h"
+#include "grd-credentials-file.h"
+#include "grd-credentials-libsecret.h"
+#include "grd-credentials-tpm.h"
 
 #define GRD_RDP_SETTINGS_SCHEMA "org.gnome.desktop.remote-desktop.rdp"
 #define GRD_VNC_SETTINGS_SCHEMA "org.gnome.desktop.remote-desktop.vnc"
 
+typedef enum _CredentialsType
+{
+  CREDENTIALS_TYPE_HEADLESS,
+  CREDENTIALS_TYPE_SCREEN_SHARE,
+} CredentialsType;
+
 typedef struct _SubCommand
 {
   const char *subcommand;
-  gboolean (* process) (int      argc,
-                        char   **argv,
-                        GError **error);
+  gboolean (* process) (CredentialsType   credentials_type,
+                        int               argc,
+                        char            **argv,
+                        GError          **error);
   int n_args;
 } SubCommand;
 
@@ -46,7 +55,8 @@ print_usage (void)
 }
 
 static int
-process_options (int                argc,
+process_options (CredentialsType    credentials_type,
+                 int                argc,
                  char             **argv,
                  const SubCommand  *subcommands,
                  int                n_subcommands)
@@ -73,7 +83,8 @@ process_options (int                argc,
           return EXIT_FAILURE;
         }
 
-      if (!subcommands[i].process (argc - 1, argv + 1, &error))
+      if (!subcommands[i].process (credentials_type,
+                                   argc - 1, argv + 1, &error))
         {
           g_printerr ("%s\n", error->message);
           return EXIT_FAILURE;
@@ -87,56 +98,89 @@ process_options (int                argc,
   return EXIT_FAILURE;
 }
 
+static GrdCredentials *
+create_credentials (CredentialsType   credentials_type,
+                    GError          **error)
+{
+  g_autoptr (GError) local_error = NULL;
+
+  switch (credentials_type)
+    {
+    case CREDENTIALS_TYPE_SCREEN_SHARE:
+      return GRD_CREDENTIALS (grd_credentials_libsecret_new ());
+    case CREDENTIALS_TYPE_HEADLESS:
+      {
+        GrdCredentialsTpm *credentials_tpm;
+        GrdCredentialsFile *credentials_file;
+        static gboolean warned_once = FALSE;
+
+        credentials_tpm = grd_credentials_tpm_new (&local_error);
+        if (credentials_tpm)
+          return GRD_CREDENTIALS (credentials_tpm);
+
+        g_debug ("Failed to create TPM credentials backend: %s", local_error->message);
+
+        if (!warned_once)
+          {
+            g_printerr ("Init TPM credentials failed, using plain text fallback.\n");
+            warned_once = TRUE;
+          }
+
+        credentials_file = grd_credentials_file_new (error);
+        if (credentials_file)
+          return GRD_CREDENTIALS (credentials_file);
+
+        return NULL;
+      }
+    }
+
+  g_assert_not_reached ();
+}
+
 #ifdef HAVE_RDP
 static gboolean
-grd_store_rdp_credentials (const char  *username,
-                           const char  *password,
-                           GError     **error)
+grd_store_rdp_credentials (CredentialsType   credentials_type,
+                           const char       *username,
+                           const char       *password,
+                           GError          **error)
 {
+  g_autoptr (GrdCredentials) credentials = NULL;
   GVariantBuilder builder;
-  g_autofree char *credentials = NULL;
 
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
   g_variant_builder_add (&builder, "{sv}",
                          "username", g_variant_new_string (username));
   g_variant_builder_add (&builder, "{sv}",
                          "password", g_variant_new_string (password));
-  credentials = g_variant_print (g_variant_builder_end (&builder), TRUE);
 
-  if (!secret_password_store_sync (GRD_RDP_CREDENTIALS_SCHEMA,
-                                   SECRET_COLLECTION_DEFAULT,
-                                   "GNOME Remote Desktop RDP credentials",
-                                   credentials,
-                                   NULL, error,
-                                   NULL))
+  credentials = create_credentials (credentials_type, error);
+  if (!credentials)
     return FALSE;
 
-  return TRUE;
+  return grd_credentials_store (credentials,
+                                GRD_CREDENTIALS_TYPE_RDP,
+                                g_variant_builder_end (&builder),
+                                error);
 }
 
 static gboolean
-grd_clear_rdp_credentials (GError **error)
+grd_clear_rdp_credentials (CredentialsType   credentials_type,
+                           GError          **error)
 {
-  g_autoptr (GError) local_error = NULL;
+  g_autoptr (GrdCredentials) credentials = NULL;
 
-  secret_password_clear_sync (GRD_RDP_CREDENTIALS_SCHEMA,
-                              NULL, &local_error,
-                              NULL);
-  if (local_error)
-    {
-      g_propagate_error (error, g_steal_pointer (&local_error));
-      return FALSE;
-    }
-  else
-    {
-      return TRUE;
-    }
+  credentials = create_credentials (credentials_type, error);
+  if (!credentials)
+    return FALSE;
+
+  return grd_credentials_clear (credentials, GRD_CREDENTIALS_TYPE_RDP, error);
 }
 
 static gboolean
-rdp_enable (int      argc,
-            char   **argv,
-            GError **error)
+rdp_enable (CredentialsType   credentials_type,
+            int               argc,
+            char            **argv,
+            GError          **error)
 {
   g_autoptr (GSettings) rdp_settings = NULL;
 
@@ -146,9 +190,10 @@ rdp_enable (int      argc,
 }
 
 static gboolean
-rdp_disable (int      argc,
-             char   **argv,
-             GError **error)
+rdp_disable (CredentialsType   credentials_type,
+             int               argc,
+             char            **argv,
+             GError          **error)
 {
   g_autoptr (GSettings) rdp_settings = NULL;
 
@@ -158,9 +203,10 @@ rdp_disable (int      argc,
 }
 
 static gboolean
-rdp_set_tls_cert (int      argc,
-                  char   **argv,
-                  GError **error)
+rdp_set_tls_cert (CredentialsType   credentials_type,
+                  int               argc,
+                  char            **argv,
+                  GError          **error)
 {
   g_autoptr (GSettings) rdp_settings = NULL;
   char *path = argv[0];
@@ -178,9 +224,10 @@ rdp_set_tls_cert (int      argc,
 }
 
 static gboolean
-rdp_set_tls_key (int      argc,
-                 char   **argv,
-                 GError **error)
+rdp_set_tls_key (CredentialsType   credentials_type,
+                 int               argc,
+                 char            **argv,
+                 GError          **error)
 {
   g_autoptr (GSettings) rdp_settings = NULL;
   char *path = argv[0];
@@ -198,25 +245,29 @@ rdp_set_tls_key (int      argc,
 }
 
 static gboolean
-rdp_set_credentials (int      argc,
-                     char   **argv,
-                     GError **error)
+rdp_set_credentials (CredentialsType   credentials_type,
+                     int               argc,
+                     char            **argv,
+                     GError          **error)
 {
-  return grd_store_rdp_credentials (argv[0], argv[1], error);
+  return grd_store_rdp_credentials (credentials_type,
+                                    argv[0], argv[1], error);
 }
 
 static gboolean
-rdp_clear_credentials (int      argc,
-                       char   **argv,
-                       GError **error)
+rdp_clear_credentials (CredentialsType   credentials_type,
+                       int               argc,
+                       char            **argv,
+                       GError          **error)
 {
-  return grd_clear_rdp_credentials (error);
+  return grd_clear_rdp_credentials (credentials_type, error);
 }
 
 static gboolean
-rdp_enable_view_only (int      argc,
-                      char   **argv,
-                      GError **error)
+rdp_enable_view_only (CredentialsType   credentials_type,
+                      int               argc,
+                      char            **argv,
+                      GError          **error)
 {
   g_autoptr (GSettings) rdp_settings = NULL;
 
@@ -226,9 +277,10 @@ rdp_enable_view_only (int      argc,
 }
 
 static gboolean
-rdp_disable_view_only (int      argc,
-                       char   **argv,
-                       GError **error)
+rdp_disable_view_only (CredentialsType   credentials_type,
+                       int               argc,
+                       char            **argv,
+                       GError          **error)
 {
   g_autoptr (GSettings) rdp_settings = NULL;
 
@@ -249,10 +301,12 @@ static const SubCommand rdp_subcommands[] = {
 };
 
 static int
-process_rdp_options (int    argc,
-                     char **argv)
+process_rdp_options (CredentialsType   credentials_type,
+                     int               argc,
+                     char            **argv)
 {
-  return process_options (argc, argv,
+  return process_options (credentials_type,
+                          argc, argv,
                           rdp_subcommands,
                           G_N_ELEMENTS (rdp_subcommands));
 }
@@ -260,9 +314,10 @@ process_rdp_options (int    argc,
 
 #ifdef HAVE_VNC
 static gboolean
-vnc_enable (int      argc,
-            char   **argv,
-            GError **error)
+vnc_enable (CredentialsType   credentials_type,
+            int               argc,
+            char            **argv,
+            GError          **error)
 {
   g_autoptr (GSettings) vnc_settings = NULL;
 
@@ -272,9 +327,10 @@ vnc_enable (int      argc,
 }
 
 static gboolean
-vnc_disable (int      argc,
-             char   **argv,
-             GError **error)
+vnc_disable (CredentialsType   credentials_type,
+             int               argc,
+             char            **argv,
+             GError          **error)
 {
   g_autoptr (GSettings) vnc_settings = NULL;
 
@@ -286,10 +342,12 @@ vnc_disable (int      argc,
 #define MAX_VNC_PASSWORD_SIZE 8
 
 static gboolean
-vnc_set_credentials (int      argc,
-                     char   **argv,
-                     GError **error)
+vnc_set_credentials (CredentialsType   credentials_type,
+                     int               argc,
+                     char            **argv,
+                     GError          **error)
 {
+  g_autoptr (GrdCredentials) credentials = NULL;
   char *password;
 
   password = argv[0];
@@ -300,42 +358,36 @@ vnc_set_credentials (int      argc,
       return FALSE;
     }
 
-  if (!secret_password_store_sync (GRD_VNC_PASSWORD_SCHEMA,
-                                   SECRET_COLLECTION_DEFAULT,
-                                   "GNOME Remote Desktop VNC password",
-                                   password,
-                                   NULL, error,
-                                   NULL))
+  credentials = create_credentials (credentials_type, error);
+  if (!credentials)
     return FALSE;
 
-  return TRUE;
+  return grd_credentials_store (credentials,
+                                GRD_CREDENTIALS_TYPE_VNC,
+                                g_variant_new_string (password),
+                                error);
 }
 
 static gboolean
-vnc_clear_credentials (int      argc,
-                       char   **argv,
-                       GError **error)
+vnc_clear_credentials (CredentialsType   credentials_type,
+                       int               argc,
+                       char            **argv,
+                       GError          **error)
 {
-  g_autoptr (GError) local_error = NULL;
+  g_autoptr (GrdCredentials) credentials = NULL;
 
-  secret_password_clear_sync (GRD_VNC_PASSWORD_SCHEMA,
-                              NULL, &local_error,
-                              NULL);
-  if (local_error)
-    {
-      g_propagate_error (error, g_steal_pointer (&local_error));
-      return FALSE;
-    }
-  else
-    {
-      return TRUE;
-    }
+  credentials = create_credentials (credentials_type, error);
+  if (!credentials)
+    return FALSE;
+
+  return grd_credentials_clear (credentials, GRD_CREDENTIALS_TYPE_VNC, error);
 }
 
 static gboolean
-vnc_set_auth_method (int      argc,
-                     char   **argv,
-                     GError **error)
+vnc_set_auth_method (CredentialsType   credentials,
+                     int               argc,
+                     char            **argv,
+                     GError          **error)
 {
   char *auth_method = argv[0];
   g_autoptr (GSettings) vnc_settings = NULL;
@@ -354,9 +406,10 @@ vnc_set_auth_method (int      argc,
 }
 
 static gboolean
-vnc_enable_view_only (int      argc,
-                      char   **argv,
-                      GError **error)
+vnc_enable_view_only (CredentialsType   credentials,
+                      int               argc,
+                      char            **argv,
+                      GError          **error)
 {
   g_autoptr (GSettings) vnc_settings = NULL;
 
@@ -366,9 +419,10 @@ vnc_enable_view_only (int      argc,
 }
 
 static gboolean
-vnc_disable_view_only (int      argc,
-                       char   **argv,
-                       GError **error)
+vnc_disable_view_only (CredentialsType   credentials,
+                       int               argc,
+                       char            **argv,
+                       GError          **error)
 {
   g_autoptr (GSettings) vnc_settings = NULL;
 
@@ -388,10 +442,12 @@ static const SubCommand vnc_subcommands[] = {
 };
 
 static int
-process_vnc_options (int    argc,
-                     char **argv)
+process_vnc_options (CredentialsType   credentials_type,
+                     int               argc,
+                     char            **argv)
 {
-  return process_options (argc, argv,
+  return process_options (credentials_type,
+                          argc, argv,
                           vnc_subcommands,
                           G_N_ELEMENTS (vnc_subcommands));
 }
@@ -444,6 +500,7 @@ print_help (void)
     _("  status [--show-credentials]                - Show current status\n"
       "\n"
       "Options:\n"
+      "  --headless                                 - Use headless credentials storage\n"
       "  --help                                     - Print this help text\n");
 
   print_usage ();
@@ -487,21 +544,30 @@ status_to_string (gboolean enabled,
 
 #ifdef HAVE_RDP
 static void
-print_rdp_status (gboolean use_colors,
-                  gboolean show_credentials)
+print_rdp_status (CredentialsType credentials_type,
+                  gboolean        use_colors,
+                  gboolean        show_credentials)
 {
+  g_autoptr (GrdCredentials) credentials = NULL;
   g_autoptr (GSettings) rdp_settings = NULL;
   g_autofree char *tls_cert = NULL;
   g_autofree char *tls_key = NULL;
-  g_autofree char *credentials_string = NULL;
-  g_autoptr (GVariant) credentials = NULL;
+  g_autoptr (GVariant) rdp_credentials = NULL;
   g_autofree char *username = NULL;
   g_autofree char *password = NULL;
   g_autoptr (GError) error = NULL;
 
-  credentials_string = secret_password_lookup_sync (GRD_RDP_CREDENTIALS_SCHEMA,
-                                                    NULL, &error,
-                                                    NULL);
+  credentials = create_credentials (credentials_type, &error);
+  if (!credentials)
+    {
+      fprintf (stderr, "Failed to initialize credential manager: %s\n",
+               error->message);
+      return;
+    }
+
+  rdp_credentials = grd_credentials_lookup (credentials,
+                                            GRD_CREDENTIALS_TYPE_RDP,
+                                            &error);
   if (error)
     {
       fprintf (stderr, "Failed to lookup RDP credentials: %s\n", error->message);
@@ -522,12 +588,10 @@ print_rdp_status (gboolean use_colors,
   printf ("\tView-only: %s\n",
           g_settings_get_boolean (rdp_settings, "view-only") ? "yes" : "no");
 
-  if (credentials_string)
+  if (rdp_credentials)
     {
-      credentials = g_variant_parse (NULL, credentials_string,
-                                     NULL, NULL, NULL);
-      g_variant_lookup (credentials, "username", "s", &username);
-      g_variant_lookup (credentials, "password", "s", &password);
+      g_variant_lookup (rdp_credentials, "username", "s", &username);
+      g_variant_lookup (rdp_credentials, "password", "s", &password);
     }
 
   if (show_credentials)
@@ -547,23 +611,37 @@ print_rdp_status (gboolean use_colors,
 
 #ifdef HAVE_VNC
 static void
-print_vnc_status (gboolean use_colors,
-                  gboolean show_credentials)
+print_vnc_status (CredentialsType credentials_type,
+                  gboolean        use_colors,
+                  gboolean        show_credentials)
 {
+  g_autoptr (GrdCredentials) credentials = NULL;
   g_autoptr (GSettings) vnc_settings = NULL;
+  g_autoptr (GVariant) password_variant = NULL;
   g_autofree char *auth_method = NULL;
-  g_autofree char *password = NULL;
+  const char *password = NULL;
   g_autoptr (GError) error = NULL;
 
-  password = secret_password_lookup_sync (GRD_VNC_PASSWORD_SCHEMA,
-                                          NULL, &error,
-                                          NULL);
+  credentials = create_credentials (credentials_type, &error);
+  if (!credentials)
+    {
+      fprintf (stderr, "Failed to initialize credential manager: %s\n",
+               error->message);
+      return;
+    }
+
+  password_variant = grd_credentials_lookup (credentials,
+                                             GRD_CREDENTIALS_TYPE_VNC,
+                                             &error);
   if (error)
     {
       fprintf (stderr, "Failed to lookup VNC credentials: %s\n", error->message);
       return;
     }
 
+  if (password_variant)
+    password = g_variant_get_string (password_variant, NULL);
+
   vnc_settings = g_settings_new (GRD_VNC_SETTINGS_SCHEMA);
 
   printf ("VNC:\n");
@@ -588,8 +666,9 @@ print_vnc_status (gboolean use_colors,
 #endif /* HAVE_VNC */
 
 static int
-print_status (int    argc,
-              char **argv)
+print_status (CredentialsType   credentials_type,
+              int               argc,
+              char            **argv)
 {
   gboolean use_colors;
   gboolean show_credentials = FALSE;
@@ -615,10 +694,10 @@ print_status (int    argc,
   use_colors = isatty (fileno (stdout));
 
 #ifdef HAVE_RDP
-  print_rdp_status (use_colors, show_credentials);
+  print_rdp_status (credentials_type, use_colors, show_credentials);
 #endif /* HAVE_RDP */
 #ifdef HAVE_VNC
-  print_vnc_status (use_colors, show_credentials);
+  print_vnc_status (credentials_type, use_colors, show_credentials);
 #endif /* HAVE_VNC */
 
   return EXIT_SUCCESS;
@@ -628,6 +707,10 @@ int
 main (int   argc,
       char *argv[])
 {
+  CredentialsType credentials_type;
+  int i;
+  int arg_shift;
+
   g_set_prgname (argv[0]);
 
   if (argc < 2)
@@ -642,22 +725,38 @@ main (int   argc,
       return EXIT_SUCCESS;
     }
 
-#ifdef HAVE_RDP
-  if (strcmp (argv[1], "rdp") == 0)
+  if (argc > 2 && strcmp (argv[1], "--headless") == 0)
     {
-      return process_rdp_options (argc - 2, argv + 2);
+      credentials_type = CREDENTIALS_TYPE_HEADLESS;
+      i = 2;
+    }
+  else
+    {
+      credentials_type = CREDENTIALS_TYPE_SCREEN_SHARE;
+      i = 1;
+    }
+
+  arg_shift = i + 1;
+
+#ifdef HAVE_RDP
+  if (strcmp (argv[i], "rdp") == 0)
+    {
+      return process_rdp_options (credentials_type,
+                                  argc - arg_shift, argv + arg_shift);
     }
 #endif
 #ifdef HAVE_VNC
-  if (strcmp (argv[1], "vnc") == 0)
+  if (strcmp (argv[i], "vnc") == 0)
     {
-      return process_vnc_options (argc - 2, argv + 2);
+      return process_vnc_options (credentials_type,
+                                  argc - arg_shift, argv + arg_shift);
     }
 #endif
 
-  if (strcmp (argv[1], "status") == 0)
+  if (strcmp (argv[i], "status") == 0)
     {
-      return print_status (argc - 2, argv + 2);
+      return print_status (credentials_type,
+                           argc - arg_shift, argv + arg_shift);
     }
 
   print_usage ();
